"""Base stack with R2 backend."""

from cdktf import S3Backend

from helicopyter import HeliStack


class BaseStack(HeliStack):
    """
    R2 backend requires the following environment variables.

    AWS_ACCESS_KEY_ID     - R2 token
    AWS_SECRET_ACCESS_KEY - R2 secret
    AWS_ENDPOINT_URL_S3   - R2 location: https://ACCOUNT_ID.r2.cloudflarestorage.com
    """

    def __init__(self, cona: str) -> None:
        super().__init__(cona)
        # Organize state files as: {cona}/{envi}/{cona}.tfstate
        backend = S3Backend(
            self,
            bucket='terraform',
            # Just in case somebody uses the default environment for two different codenames
            key=f'{cona}.tfstate',
            region='auto',
            workspace_key_prefix=cona,
        )
        # Sadly `attribute=True` above was resulting in `attribute = undefined` in the HCL
        for attribute in (
            'skip_credentials_validation',
            'skip_metadata_api_check',
            'skip_region_validation',
            'skip_requesting_account_id',
            'skip_s3_checksum',
            'use_path_style',
        ):
            backend.add_override(attribute, 'true')



#!/usr/bin/env python
from constructs import Construct
from cdk8s import App, Chart


class MyChart(Chart):
    def __init__(self, scope: Construct, id: str):
        super().__init__(scope, id)

        # define resources here


app = App()
MyChart(app, "{{ $base }}")

app.synth()



"""Generate Hashicorp Configuration Language (HCL) or JSON from Python."""

from collections.abc import Iterable
from importlib import import_module
from json import dump
from pathlib import Path
from re import sub
from subprocess import check_output
from typing import Any, TypeVar

from cdk8s import Chart
from cdktf import App, TerraformElement, TerraformStack
from constructs import Construct, Node
from tap import Tap


class HeliChart(Chart):
    def __init__(self, cona: str) -> None:
        super().__init__(App(), cona)
        self.cona = cona


class HeliStack(TerraformStack):
    def __init__(self, cona: str) -> None:
        # Something is automatically creating outdir, which is cdktf.out by default
        super().__init__(App(outdir='.'), cona)

        self.cona = cona
        self.imports: dict[str, str] = {}  # {to_0: id_0, to_1, id_1, ...}
        self._scopes: dict[str, Construct] = {}

    def _allocate_logical_id(self, element: Node | TerraformElement) -> str:
        if isinstance(element, Node):
            # Mostly for mypy. Patches to support AWS CDK welcome.
            raise TypeError('AWS CDK unsupported; please use CDKTF')
        return element.node.id

    def override(self, **kwargs: Any) -> None:
        """Call add_override for each keyword argument."""
        for key, value in kwargs.items():
            self.add_override(key, value)

    def provide(self, name: str, **kwargs: Any) -> type[TerraformElement]:
        """
        Return a Provider class instance given its short name.

        Example usage:
        stack.provide('github', owner='biobuddies')
        """
        module = import_module(f'cdktf_cdktf_provider_{name}.provider')
        return getattr(module, f'{name.title()}Provider')(self, 'this', **kwargs)

    E = TypeVar('E', bound=TerraformElement)

    def push(
        self,
        Element: type[E],  # noqa: N803
        id_: str,
        *args: Any,
        import_id: str = '',
        **kwargs: Any,
    ) -> E:
        """
        Return new instance of Element (data, local, output, resource, or variable).

        In contrast to running Element(...) standalone, the new instance will be named in the
        traditional Terraform style.

        If an import_id is provided, save it to self.imports for use by multisynth().

        Example usage:
        from cdktf_cdktf_provider_cloudflare.zero_trust_access_application import (
            ZeroTrustAccessApplication
        )
        stack.push(ZeroTrustAccessApplication, 'mydomain-wildcard', domain='*.mydomain.com')
        """
        if Element.__module__ == 'cdktf':
            scope_name = Element.__name__.lower().replace('terraform', '')
        else:
            scope_name = Element.__module__.replace('cdktf_cdktf_provider_', '').replace('.', '_')
        if scope_name not in self._scopes:
            self._scopes[scope_name] = Construct(self, scope_name)

        print(f'Pushing {scope_name}.{id_}')
        element = Element(self._scopes[scope_name], id_, *args, **kwargs)
        if import_id:
            self.imports[
                f'{Element.__module__.replace("cdktf_cdktf_provider_", "").replace(".", "_")}.{id_}'
            ] = import_id
        return element


# ruff: noqa: T201
def multisynth(
    all_or_conas_or_paths: Iterable[str],
    *,
    change_directory: Path | None,
    hashicorp_configuration_language: bool,
    format_with: str,
) -> None:
    """
    Generate Hashicorp Configuration Language (HCL) or JSON.

    An import block will generated for each (key, value) pair in stack.imports.
    """
    if not all_or_conas_or_paths:
        print('No codenames specified. Doing nothing.')
        return

    top_directory = change_directory if change_directory else Path.cwd()

    if 'all' in all_or_conas_or_paths or 'helicopyter.py' in all_or_conas_or_paths:
        conas_or_paths = {
            file.parent.parent.name
            for file in (top_directory / 'deploys').glob('**/terraform/main.py')
        }
    else:
        conas_or_paths = set(all_or_conas_or_paths)

    for cona_or_path in conas_or_paths:
        path_to_check = Path(cona_or_path)
        if (
            path_to_check.exists()
            and path_to_check.name == 'main.py'
            and path_to_check.parent.name == 'terraform'
        ):
            cona = path_to_check.parent.parent.name
        else:
            cona = cona_or_path
        relative_path = (
            f'deploys/{cona}/terraform/main.tf{"" if hashicorp_configuration_language else ".json"}'
        )
        print(f'Generating {relative_path}')
        try:
            module_path = f'deploys.{cona}.terraform.main'
            main = import_module(module_path)
            stack = main.synth.__annotations__['stack'](cona)
            main.synth(stack)
        except (AttributeError, ImportError, KeyError, TypeError):
            python_file = module_path.replace('.', '/') + '.py'
            print(f'`def synth(stack: HeliStack):` appears to be missing from {python_file}')
            raise
        if hashicorp_configuration_language:
            top = '# AUTOGENERATED by helicopyter\n\n'
            if stack.imports:
                top += (
                    '\n\n'.join(
                        'import {\n  to = %s\n  id = "%s"\n}' % (resource, import_id)  # noqa: UP031
                        for resource, import_id in stack.imports.items()
                    )
                    + '\n\n'
                )
            unformatted = top + stack.to_hcl_terraform()['hcl']
            autoformatted = check_output(  # noqa: S603
                [format_with, 'fmt', '-'], input=unformatted.encode()
            ).decode()
            formatted = sub(
                r'\n{3,}',
                '\n\n',
                autoformatted.replace('}\n\n\n}', '}\n}').replace('}\nresource', '}\n\nresource'),
            )
            (top_directory / relative_path).write_text(formatted.strip() + '\n')
        else:
            dictionary = stack.to_terraform()
            dictionary['//']['AUTOGENERATED'] = 'by helicopyter'
            if stack.imports:
                dictionary['import'] = [
                    {'to': resource, 'id': import_id}
                    for resource, import_id in stack.imports.items()
                ]
            with (top_directory / relative_path).open('w') as output:
                dump(dictionary, output, indent=4, sort_keys=True)
                output.write('\n')


class Parameters(Tap):
    conas: list[str]  # space-separated COdeNAmes
    directory: Path | None = None
    format_with: str = 'terraform'
    hashicorp_configuration_language: bool = True

    def configure(self) -> None:  # noqa: D102
        self.add_argument('conas')  # Positional argument
        self.add_argument('-C', '--directory')  # Like make and tar



"""Codify Allowedflare infrastructure."""

from os import environ

from cdktf_cdktf_provider_cloudflare.workers_route import WorkersRoute
from cdktf_cdktf_provider_cloudflare.zero_trust_access_application import ZeroTrustAccessApplication
from cdktf_cdktf_provider_cloudflare.zero_trust_access_identity_provider import (
    ZeroTrustAccessIdentityProvider,
)
from cdktf_cdktf_provider_cloudflare.zero_trust_access_policy import (
    ZeroTrustAccessPolicy,
    ZeroTrustAccessPolicyInclude,
)

from stacks.base import BaseStack


def synth(stack: BaseStack) -> None:
    stack.provide('cloudflare')

    # If this was a private repository, I'd probably set these variables using string literals
    account_id = environ['CLOUDFLARE_ACCOUNT_ID']
    email = environ['ALLOWEDFLARE_EMAIL']
    private_domain = environ['ALLOWEDFLARE_PRIVATE_DOMAIN']
    zone_id = environ['CLOUDFLARE_ZONE_ID']

    stack.push(
        ZeroTrustAccessIdentityProvider,
        'this',
        account_id=account_id,
        name='One-time PIN',
        type='onetimepin',
    )

    stack.push(
        ZeroTrustAccessApplication,
        'this',
        domain=f'*.{private_domain}',
        session_duration='24h',
        skip_interstitial=True,
        policies=[
            stack.push(
                ZeroTrustAccessPolicy,
                'email-in',
                account_id=account_id,
                name='email-in',
                decision='allow',
                include=[ZeroTrustAccessPolicyInclude(email=[email])],
            ).id,
            stack.push(
                ZeroTrustAccessPolicy,
                'service-in',
                account_id=account_id,
                name='service-in',
                decision='non_identity',
                include=[ZeroTrustAccessPolicyInclude(any_valid_service_token=True)],
            ).id,
        ],
    )

    stack.push(
        WorkersRoute,
        'this',
        pattern=f'*.{private_domain}/x/*',
        script_name='allowedflare-proxy',  # script in v5
        zone_id=zone_id,
    )



"""Generate Hashicorp Configuration Language (HCL) or JSON from Python."""

from helicopyter import Parameters, multisynth

args = Parameters().parse_args()
multisynth(
    args.conas,
    change_directory=args.directory,
    format_with=args.format_with,
    hashicorp_configuration_language=args.hashicorp_configuration_language,
)



"""
Avoid git URLs breaking PyPI uploads.

Similar problem to:
https://stackoverflow.com/questions/54887301/how-can-i-use-git-repos-as-dependencies-for-my-pypi-package
"""

from io import StringIO
from typing import TextIO

import setuptools
from packaging.metadata import Metadata
from setuptools._core_metadata import _write_requirements  # type: ignore[import-not-found]
from setuptools.build_meta import *  # noqa: F403


def write_pypi_compatible_requirements(self: Metadata, final_file: TextIO) -> None:
    """Mark requirements with URLs as external."""
    initial_file = StringIO()
    _write_requirements(self, initial_file)
    initial_file.seek(0)
    for initial_line in initial_file:
        final_line = initial_line
        metadata = Metadata.from_email(initial_line, validate=False)
        if metadata.requires_dist and metadata.requires_dist[0].url:
            final_line = initial_line.replace('Requires-Dist:', 'Requires-External:')
        final_file.write(final_line)


setuptools._core_metadata._write_requirements = write_pypi_compatible_requirements
